/*******************************************************************************
 * Copyright 2015 Julian Oppermann, TU Darmstadt
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

#include "llvm/IR/LLVMContext.h"
#include "llvm/IR/Value.h"
#include "llvm/IR/Instructions.h"
#include "llvm/IR/Constants.h"
#include "llvm/IR/Function.h"
#include "llvm/IR/Argument.h"
#include "llvm/IR/BasicBlock.h"
#include "llvm/IR/Dominators.h"
#include "llvm/IR/Intrinsics.h"
#include "llvm/IR/PatternMatch.h"
#include "llvm/IR/CFG.h"

#include <vector>
#include <set>
#include <map>
#include <deque>
#include <algorithm>

#include "Equations.h"
#include "EquationBuilder.h"
#include "config.h"

using namespace llvm;
using namespace llvm::PatternMatch;
using namespace std;

namespace cellml {
namespace opt {

void EquationBuilder::build(Function *EF) {
	DT.recalculate(*EF);
	visit(EF);
}

void EquationBuilder::visitBinaryOperator(BinaryOperator &I) {
	/* special case: LLVM doesn't have a NEG instruction */
	if (BinaryOperator::isFNeg(&I)) {
		Expr *rhs = getExpr(I.getOperand(1));
		UnaryExpr *neg = new UnaryExpr(UnaryFunction::neg, rhs);
		insertExpr(&I, neg);
		return;
	}

	Expr *lhs = getExpr(I.getOperand(0));
	Expr *rhs = getExpr(I.getOperand(1));

	BinaryFunction func;
	switch (I.getOpcode()) {
	case BinaryOperator::BinaryOps::FAdd: func = BinaryFunction::add; break;
	case BinaryOperator::BinaryOps::FSub: func = BinaryFunction::sub; break;
	case BinaryOperator::BinaryOps::FMul: func = BinaryFunction::mul; break;
	case BinaryOperator::BinaryOps::FDiv: func = BinaryFunction::div; break;
	case BinaryOperator::BinaryOps::And:  func = BinaryFunction::land; break;
#ifndef CELLML_TO_VHDL_COMPAT_MODE
	case BinaryOperator::BinaryOps::Or:   func = BinaryFunction::lor; break;
#endif
	default:
		assert (false && "unhandled binary operator");
	}

	BinaryExpr *binaryExpr = new BinaryExpr(func, lhs, rhs);
	insertExpr(&I, binaryExpr);
}

void EquationBuilder::visitICmp(ICmpInst &I) {
	/* handle the implicit 'something != 0' that is inserted when writing 'if (something)'
	 * as the rest of the calculation are floating point, we don't expect any other
	 * integer compares */
	if (I.getPredicate() == ICmpInst::Predicate::ICMP_NE) {
		if (ConstantInt *C = dyn_cast<ConstantInt>(I.getOperand(1))) {
			if (C->isZeroValue()) {
				insertExpr(&I, getExpr(I.getOperand(0)));
				return;
			}
		}
	}

	assert (false && "unexpected ICmp instruction");
}

void EquationBuilder::visitZExt(ZExtInst &I) {
	/* the zero extension will happen implicitly in the C code, so just
	 * pass through the expression */
	insertExpr(&I, getExpr(I.getOperand(0)));
}

void EquationBuilder::visitFCmp(FCmpInst &I) {
	Expr *lhs = getExpr(I.getOperand(0));
	Expr *rhs = getExpr(I.getOperand(1));

	BinaryFunction func;
	switch (I.getPredicate()) {
		case FCmpInst::Predicate::FCMP_ULT:
		case FCmpInst::Predicate::FCMP_OLT: func = BinaryFunction::lt; break;
		case FCmpInst::Predicate::FCMP_ULE:
		case FCmpInst::Predicate::FCMP_OLE: func = BinaryFunction::le; break;
		case FCmpInst::Predicate::FCMP_UGT:
		case FCmpInst::Predicate::FCMP_OGT: func = BinaryFunction::gt; break;
		case FCmpInst::Predicate::FCMP_UGE:
		case FCmpInst::Predicate::FCMP_OGE: func = BinaryFunction::ge; break;
#ifndef CELLML_TO_VHDL_COMPAT_MODE
		case FCmpInst::Predicate::FCMP_UEQ:
		case FCmpInst::Predicate::FCMP_OEQ: func = BinaryFunction::eq; break;
		case FCmpInst::Predicate::FCMP_UNE:
		case FCmpInst::Predicate::FCMP_ONE: func = BinaryFunction::ne; break;
#endif
		default:
			assert (false && "unhandled compare predicate");
		}

	BinaryExpr *cmp = new BinaryExpr(func, lhs, rhs);
	insertExpr(&I, cmp);
}

void EquationBuilder::handleCallToUnaryFunction(CallInst &I, UnaryFunction function) {
	assert (I.getNumArgOperands() == 1);
	Expr *op = getExpr(I.getArgOperand(0));

	UnaryExpr *unary = new UnaryExpr(function, op);
	insertExpr(&I, unary);
}

void EquationBuilder::handleCallToBinaryFunction(CallInst &I, BinaryFunction function) {
	assert (I.getNumArgOperands() == 2);
	Expr *lhs = getExpr(I.getArgOperand(0));
	Expr *rhs = getExpr(I.getArgOperand(1));

	BinaryExpr *binary = new BinaryExpr(function, lhs, rhs);
	insertExpr(&I, binary);
}

void EquationBuilder::visitCall(CallInst &I) {
	Function *callee = I.getCalledFunction();
	assert (callee && callee->isIntrinsic() && "call target unknown");

	switch (callee->getIntrinsicID()) {
	case Intrinsic::pow:   handleCallToBinaryFunction(I, BinaryFunction::pow); break;
	case Intrinsic::exp:   handleCallToUnaryFunction(I, UnaryFunction::exp);   break;
	case Intrinsic::log:   handleCallToUnaryFunction(I, UnaryFunction::log);   break;
	case Intrinsic::floor: handleCallToUnaryFunction(I, UnaryFunction::floor); break;
	case Intrinsic::fabs:  handleCallToUnaryFunction(I, UnaryFunction::abs);   break;
	default:
		assert (false && "unexpected function");
	}
}

Expr* EquationBuilder::handlePhiPredBlocks(BasicBlock *branchBlock, PHINode *phi) {
	BranchInst *br = dyn_cast<BranchInst>(branchBlock->getTerminator());
	assert (br && br->isConditional() && "unhandled CFG structure");

	BasicBlock *phiBlock  = phi->getParent();
	BasicBlock *trueSucc  = br->getSuccessor(0);
	BasicBlock *falseSucc = br->getSuccessor(1);
	assert (trueSucc != falseSucc);

	set<BasicBlock *> trueBlocks, falseBlocks;

	if (trueSucc == phiBlock)
		trueBlocks.insert(branchBlock);
	if (falseSucc == phiBlock)
		falseBlocks.insert(branchBlock);

	for (auto BI = phi->block_begin(), BE = phi->block_end(); BI != BE; BI++) {
		BasicBlock *pred = *BI;
		if (DT.dominates(trueSucc, pred))
			trueBlocks.insert(pred);
		else if (DT.dominates(falseSucc, pred))
			falseBlocks.insert(pred);
		// else: ignore, it's a phi pred from a "higher up" branch
	}
	assert (!trueBlocks.empty() && !falseBlocks.empty());

	Expr *condExpr, *trueExpr, *falseExpr;
	condExpr = getExpr(br->getCondition());

	if (trueBlocks.size() == 1)
		trueExpr = getExpr(phi->getIncomingValueForBlock(*trueBlocks.begin()));
	else
		trueExpr = handlePhiPredBlocks(trueSucc, phi);

	if (falseBlocks.size() == 1)
		falseExpr = getExpr(phi->getIncomingValueForBlock(*falseBlocks.begin()));
	else
		falseExpr = handlePhiPredBlocks(falseSucc, phi);

	SelectExpr *select = new SelectExpr(condExpr, trueExpr, falseExpr);
	insertIntermediateExpr(select);

	return select;
}

void EquationBuilder::visitPHI(PHINode &I) {
	assert (I.getNumIncomingValues() >= 2);

	BasicBlock *phiBlock = I.getParent();
	BasicBlock *idom     = DT.getNode(phiBlock)->getIDom()->getBlock();

	Expr *select = handlePhiPredBlocks(idom, &I);

	insertExpr(&I, select);
}

void EquationBuilder::visitSelect(SelectInst &I) {
	Expr *cond = getExpr(I.getCondition());
	Expr *ifTrue = getExpr(I.getTrueValue());
	Expr *ifFalse = getExpr(I.getFalseValue());

	SelectExpr *select = new SelectExpr(cond, ifTrue, ifFalse);
	insertExpr(&I, select);
}

VarExpr *EquationBuilder::handleAddress(Value *addr) {
	if (GetElementPtrInst *gep = dyn_cast<GetElementPtrInst>(addr)) {
		Argument *base = dyn_cast<Argument>(gep->getPointerOperand());
		assert (base);

		assert (gep->getNumIndices() == 1);
		ConstantInt *index = dyn_cast<ConstantInt>(gep->idx_begin());
		assert (index);

		VarExpr *var = new VarExpr(base, index);
		insertExpr(addr, var);
		return var;
	}
	else if (Argument *base = dyn_cast<Argument>(addr)) {
		LLVMContext &C = getGlobalContext();
		ConstantInt *O = cast<ConstantInt>(ConstantInt::getNullValue(Type::getInt32Ty(C)));
		VarExpr *var = new VarExpr(base, O);
		insertExpr(addr, var);
		return var;
	}

	assert (false && "unsupported address");
	return nullptr;
}

void EquationBuilder::visitLoad(LoadInst &I) {
	Value   *addr = I.getPointerOperand();
	VarExpr *var  = handleAddress(addr);
	insertExpr(&I, var);
}

void EquationBuilder::visitStore(StoreInst &I) {
	VarExpr *var  = handleAddress(I.getPointerOperand());
	Expr    *expr = getExpr(I.getValueOperand());

	equationSystem.addEquation(var, expr, equationSystem.end());
}

void EquationBuilder::insertExpr(Value *V, Expr *expr) {
	expr->setNumUses(V->getNumUses());

	equationSystem.addExpr(expr);
	symbolTable.insert(std::make_pair(V, expr));
}

void EquationBuilder::insertIntermediateExpr(Expr *expr) {
	expr->setNumUses(1);
	equationSystem.addExpr(expr);
}

Expr* EquationBuilder::getExpr(Value *V) {
	if (symbolTable.count(V) == 0) {
		/* handle special cases here on-the-fly */
		if (ConstantFP *constant = dyn_cast<ConstantFP>(V)) {
			ConstExpr *constExpr = new ConstExpr(constant);
			insertExpr(V, constExpr);
			return constExpr;
		}

		if (Argument *arg = dyn_cast<Argument>(V)) {
			if (arg->getName().equals("VOI")) {
				VarExpr *voi = new VarExpr(arg, nullptr);
				insertExpr(V, voi);
				return voi;
			}
		}

		if (ConstantExpr *constantExpr = dyn_cast<ConstantExpr>(V)) {
			/* this is a weird pattern that sometimes arises */
			if (match(constantExpr, m_FAdd(m_FMul(m_Undef(), m_SpecificFP(0.0)), m_FMul(m_Undef(), m_SpecificFP(0.0))))) {
				ConstExpr *constExpr = new ConstExpr(cast<ConstantFP>(ConstantFP::getNullValue(constantExpr->getType())));
				insertExpr(V, constExpr);
				return constExpr;
			}
		}

		V->dump();
		assert (false && "unknown value");
	}

	return symbolTable.find(V)->second;
}

}
}
