/*******************************************************************************
 * Copyright 2015 Julian Oppermann, TU Darmstadt
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

#ifndef EQUATIONS_H_
#define EQUATIONS_H_

#include "llvm/IR/Value.h"
#include "llvm/IR/Constants.h"
#include "llvm/IR/Argument.h"

#include <list>
#include <set>
#include <map>

#include "config.h"

using namespace llvm;

namespace cellml {
namespace opt {

class Expr;
class VarExpr;
class TmpVarExpr;
class ConstExpr;
class UnaryExpr;
class BinaryExpr;
class SelectExpr;
class Equation;

typedef std::set<Expr *>          ExprSet;
typedef std::list<Expr *>         ExprList;
typedef std::list<Equation *>     EquationList;
typedef std::map<Value *, Expr *> SymbolTable;

typedef enum { Var, TmpVar, Const, Unary, Binary, Select }    ExprKind;
typedef enum { exp, log, floor, abs, neg }                    UnaryFunction;

#ifdef CELLML_TO_VHDL_COMPAT_MODE
typedef enum { add, sub, mul, div, pow, lt, le, gt, ge, land} BinaryFunction;
#else
typedef enum { add, sub, mul, div, pow, lt, le, gt, ge, eq, ne, land, lor} BinaryFunction;
#endif

class ExprVisitor {
public:
	virtual void visit(VarExpr *var)       = 0;
	virtual void visit(TmpVarExpr *tmp)    = 0;
	virtual void visit(ConstExpr *konst)   = 0;
	virtual void visit(UnaryExpr *unary)   = 0;
	virtual void visit(BinaryExpr *binay)  = 0;
	virtual void visit(SelectExpr *select) = 0;

	virtual ~ExprVisitor() {}
};

class Expr {
	friend class EquationSystem;
	static unsigned lastId;

	ExprKind kind;
	unsigned id;
	unsigned numUses   = 0;
	Expr    *substExpr = nullptr;

public:
	Expr(ExprKind kind) : kind(kind), id(++lastId) { }

	ExprKind getKind() const {
		return kind;
	}

	unsigned getId() const {
		return id;
	}

	unsigned getNumUses() const {
		return numUses;
	}

	void setNumUses(unsigned numUses) {
		this->numUses = numUses;
	}

	Expr *getSubstExpr() const {
		return substExpr;
	}

	void setSubstExpr(Expr *substExpr) {
		this->substExpr = substExpr;
	}

	virtual void accept(ExprVisitor *visitor) = 0;

	virtual ~Expr() {}
};

class VarExpr : public Expr {
	Argument    *base;
	ConstantInt *index;

public:
	VarExpr(Argument *base, ConstantInt *index) : Expr(ExprKind::Var), base(base), index(index) {}

	Argument* getBase() const {
		return base;
	}

	ConstantInt* getIndex() const {
		return index;
	}

	void accept(ExprVisitor *visitor) {
		visitor->visit(this);
	}
};

class TmpVarExpr : public Expr {
	unsigned tmpId;

public:
	TmpVarExpr(unsigned tmpId) : Expr(ExprKind::TmpVar), tmpId(tmpId) {}

	unsigned getTmpId() const {
		return tmpId;
	}

	void accept(ExprVisitor *visitor) {
		visitor->visit(this);
	}
};

class ConstExpr : public Expr {
	ConstantFP *floatValue;

public:
	ConstExpr(ConstantFP *floatValue) : Expr(ExprKind::Const), floatValue(floatValue) {}

	ConstantFP *getFloatValue() const {
		return floatValue;
	}

	void accept(ExprVisitor *visitor) {
		visitor->visit(this);
	}
};

class UnaryExpr : public Expr {
	UnaryFunction function;
	Expr         *op;

public:
	UnaryExpr(UnaryFunction function, Expr *op) : Expr(ExprKind::Unary), function(function), op(op) {}

	UnaryFunction getFunction() const {
		return function;
	}

	Expr *getOp() const {
		return op;
	}

	void accept(ExprVisitor *visitor) {
		visitor->visit(this);
	}
};

class BinaryExpr : public Expr {
	BinaryFunction function;
	Expr          *lhs, *rhs;

public:
	BinaryExpr(BinaryFunction function, Expr *lhs, Expr *rhs) : Expr(ExprKind::Binary), function(function), lhs(lhs), rhs(rhs) {}

	BinaryFunction getFunction() const {
		return function;
	}

	Expr *getLhs() const {
		return lhs;
	}

	Expr *getRhs() const {
		return rhs;
	}

	void accept(ExprVisitor *visitor) {
		visitor->visit(this);
	}
};

class SelectExpr : public Expr {
	Expr *cond, *ifTrue, *ifFalse;

public:
	SelectExpr(Expr *cond, Expr *ifTrue, Expr *ifFalse) : Expr(ExprKind::Select), cond(cond), ifTrue(ifTrue), ifFalse(ifFalse) {}

	Expr *getCond() const {
		return cond;
	}

	Expr *getIfTrue() const {
		return ifTrue;
	}

	Expr *getIfFalse() const {
		return ifFalse;
	}

	void accept(ExprVisitor *visitor) {
		visitor->visit(this);
	}
};

class Equation {
	friend class EquationSystem;
	static unsigned lastId;

	unsigned id;
	Expr    *lhs;
	Expr    *rhs;

public:
	Equation(Expr *lhs, Expr *rhs) : id(++lastId), lhs(lhs), rhs(rhs) {}

	unsigned getId() const {
		return id;
	}

	Expr *getLhs() const {
		return lhs;
	}

	Expr *getRhs() const {
		return rhs;
	}
};

}
}

#endif /* EQUATIONS_H_ */
