/*******************************************************************************
 * Copyright 2015 Julian Oppermann, TU Darmstadt
 *                Ting Yu, University of Auckland
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

#include "llvm/Pass.h"
#include "llvm/IR/Function.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Support/FileSystem.h"
#include "llvm/ADT/Statistic.h"

#include "Equations.h"
#include "EquationSystem.h"
#include "InstanceInfo.h"
#include "config.h"

#include <string>
#include <sstream>

#define DEBUG_TYPE "estimatearea"

#define CHECK_VISITED(x) if (! visited.insert(x).second) return;

using namespace llvm;

namespace cellml {
namespace opt {

STATISTIC(NumAdd, "A) Number of additions");
STATISTIC(NumSub, "B) Number of subtractions");
STATISTIC(NumMul, "C) Number of multiplications");
STATISTIC(NumDiv, "D) Number of divisions");
STATISTIC(NumPow, "E) Number of powers");
STATISTIC(NumExp, "F) Number of exponentations");
STATISTIC(NumLog, "G) Number of logarithms");
STATISTIC(NumFlr, "H) Number of floors");
STATISTIC(NumAbs, "I) Number of absolute values");
STATISTIC(NumNeg, "J) Number of negations");
STATISTIC(NumCmp, "K) Number of comparisons");
STATISTIC(NumBit, "L) Number of logical operations");

STATISTIC(NumLUT, "M) est. number of Adaptive Look-Up Tables");
STATISTIC(NumDLR, "N) est. number of Dedicated Logic Registers");
STATISTIC(NumALM, "O) est. number of Adaptive Logic Modules");
STATISTIC(NumDSP, "P) est. number of 18-bit DSPs");

class EstimateArea : public FunctionPass, ExprVisitor {
	ExprSet visited;

	void addFPAdd()  { ++NumAdd; NumLUT += 269;  NumDLR += 622;  NumALM += 395;  NumDSP += 0;  }
	void addFPSub()  { ++NumSub; NumLUT += 269;  NumDLR += 622;  NumALM += 395;  NumDSP += 0;  } // implemented as add
	void addFPMul()  { ++NumMul; NumLUT += 73;   NumDLR += 219;  NumALM += 132;  NumDSP += 4;  }
	void addFPDiv()  { ++NumDiv; NumLUT += 1188; NumDLR += 1407; NumALM += 1116; NumDSP += 0;  }
	void addFPPow()  { ++NumPow; NumLUT += 1808; NumDLR += 3307; NumALM += 2058; NumDSP += 31; }
	void addFPExp()  { ++NumExp; NumLUT += 436;  NumDLR += 878;  NumALM += 507;  NumDSP += 2;  }
	void addFPLog()  { ++NumLog; NumLUT += 831;  NumDLR += 1210; NumALM += 808;  NumDSP += 18; }
	void addFPFlr()  { ++NumFlr; /* cheap hardware implementation */                           }
	void addFPAbs()  { ++NumAbs; /* cheap hardware implementation */                           }
	void addFPNeg()  { ++NumNeg; NumLUT += 269;  NumDLR += 622;  NumALM += 395;  NumDSP += 0;  } // implemented as 0-x
	void addFPCmp()  { ++NumCmp; NumLUT += 269;  NumDLR += 622;  NumALM += 395;  NumDSP += 0;  } // implemented as sub
	void addBitOp()  { ++NumBit; /* cheap hardware implementation */                           }

public:
	static char ID;
	EstimateArea() : FunctionPass(ID) {}

	bool runOnFunction(Function &F) {
		EquationSystem &equationSystem = getAnalysis<EquationSystem>();

		visited.clear();
		for (Equation *eq : equationSystem) {
			eq->getRhs()->accept(this);
		}

		InstanceInfo &II = getAnalysis<InstanceInfo>();

		std::ostringstream filename;
		filename << "stats." << II.OptimisationLevelName.str() << ".csv";
		std::string error;
		raw_fd_ostream StatFile(filename.str().c_str(), error, sys::fs::F_None);

		StatFile << II.ModelName << ',';
		StatFile << II.OptimisationLevelName << ',';
		StatFile << NumAdd << ',';
		StatFile << NumSub << ',';
		StatFile << NumMul << ',';
		StatFile << NumDiv << ',';
		StatFile << NumPow << ',';
		StatFile << NumExp << ',';
		StatFile << NumLog << ',';
		StatFile << NumFlr << ',';
		StatFile << NumAbs << ',';
		StatFile << NumNeg << ',';
		StatFile << NumCmp << ',';
		StatFile << NumBit << ',';
		StatFile << NumLUT << ',';
		StatFile << NumDLR << ',';
		StatFile << NumALM << ',';
		StatFile << NumDSP << '\n';

		StatFile.close();

		return false;
	}

	void getAnalysisUsage(AnalysisUsage &AU) const {
		AU.addRequired<EquationSystem>();
		AU.addRequired<InstanceInfo>();
		AU.setPreservesAll();
	}

	void visit(VarExpr *var) {
		// do nothing
	}

	void visit(TmpVarExpr *tmp) {
		// do nothing
	}

	void visit(ConstExpr *konst) {
		// do nothing
	}

	void visit(UnaryExpr *unary) {
		CHECK_VISITED(unary);

		switch (unary->getFunction()) {
		case UnaryFunction::exp:   addFPExp(); break;
		case UnaryFunction::log:   addFPLog(); break;
		case UnaryFunction::floor: addFPFlr(); break;
		case UnaryFunction::abs:   addFPAbs(); break;
		case UnaryFunction::neg:   addFPNeg(); break;
		default:
			assert (false && "Unhandled unary function");
		}

		unary->getOp()->accept(this);
	}

	void visit(BinaryExpr *binary) {
		CHECK_VISITED(binary);

		switch (binary->getFunction()) {
		case BinaryFunction::add:  addFPAdd(); break;
		case BinaryFunction::sub:  addFPSub(); break;
		case BinaryFunction::mul:  addFPMul(); break;
		case BinaryFunction::div:  addFPDiv(); break;
		case BinaryFunction::pow:  addFPPow(); break;
		case BinaryFunction::lt:
		case BinaryFunction::le:
		case BinaryFunction::gt:
		case BinaryFunction::ge:   addFPCmp(); break;
		case BinaryFunction::land: addBitOp(); break;
#ifndef CELLML_TO_VHDL_COMPAT_MODE
		case BinaryFunction::eq:
		case BinaryFunction::ne:   addFPCmp(); break;
		case BinaryFunction::lor:  addBitOp(); break;
#endif
		default:
			assert (false && "Unhandled binary function");
		}

		binary->getLhs()->accept(this);
		binary->getRhs()->accept(this);
	}

	void visit(SelectExpr *select) {
		CHECK_VISITED(select);

		select->getCond()->accept(this);
		select->getIfTrue()->accept(this);
		select->getIfFalse()->accept(this);
	}

};

char EstimateArea::ID = 0;

Pass *createAreaEstimaterPass() {
	return new EstimateArea();
}

}
}
