/*******************************************************************************
 * Copyright 2015 Julian Oppermann, TU Darmstadt
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

#include "llvm/Pass.h"
#include "llvm/IR/Function.h"
#include "llvm/IR/Instructions.h"
#include "llvm/IR/InstIterator.h"
#include "llvm/IR/Intrinsics.h"
#include "llvm/Target/TargetLibraryInfo.h"

#include "CustomPasses.h"

using namespace llvm;

namespace cellml {
namespace opt {

class MapMathFunctionsToIntrinsics : public FunctionPass {
public:
	static char ID;
	MapMathFunctionsToIntrinsics() : FunctionPass(ID) {}

	bool runOnFunction(Function &F) {
		bool changed = false;

		Module *M = F.getParent();
		TargetLibraryInfo &TLI = getAnalysis<TargetLibraryInfo>();

		for (auto II = inst_begin(F), IE = inst_end(F); II != IE; II++) {
			if (CallInst *call = dyn_cast<CallInst>(&*II)) {
				if (Function *callee = call->getCalledFunction()) {
					LibFunc::Func funcId;
					if (TLI.getLibFunc(callee->getName(), funcId)) {
						Function *intr = nullptr;
						ArrayRef<Type *> doubleTyArr(Type::getDoubleTy(F.getContext()));

						switch (funcId) {
						case LibFunc::pow:   intr = Intrinsic::getDeclaration(M, Intrinsic::pow,   doubleTyArr); break;
						case LibFunc::exp:   intr = Intrinsic::getDeclaration(M, Intrinsic::exp,   doubleTyArr); break;
						case LibFunc::log:   intr = Intrinsic::getDeclaration(M, Intrinsic::log,   doubleTyArr); break;
						case LibFunc::floor: intr = Intrinsic::getDeclaration(M, Intrinsic::floor, doubleTyArr); break;
						case LibFunc::fabs:  intr = Intrinsic::getDeclaration(M, Intrinsic::fabs,  doubleTyArr); break;
						default:
							assert (false && "unexpected function");
						}

						call->setCalledFunction(intr);
						changed = true;
					}
				}
			}
		}

		return changed;
	}

	void getAnalysisUsage(AnalysisUsage &AU) const {
		AU.addRequired<TargetLibraryInfo>();
		AU.setPreservesCFG();
	}
};

char MapMathFunctionsToIntrinsics::ID = 0;

Pass *createIntrinsicMapperPass() {
	return new MapMathFunctionsToIntrinsics();
}

}
}
