/*******************************************************************************
 * Copyright 2015 Julian Oppermann, TU Darmstadt
 *                Ting Yu, University of Auckland
 *                Oliver Sinnen, University of Auckland
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

#include "llvm/Pass.h"
#include "llvm/IR/Function.h"
#include "llvm/IR/Intrinsics.h"
#include "llvm/IR/Constants.h"
#include "llvm/IR/InstIterator.h"
#include "llvm/IR/PatternMatch.h"
#include "llvm/IR/Dominators.h"
#include "llvm/ADT/APFloat.h"
#include "llvm/ADT/APSInt.h"
#include "llvm/ADT/SmallVector.h"
#include "llvm/ADT/Statistic.h"
#include "llvm/Support/MathExtras.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Support/Format.h"
#include "llvm/Support/FileSystem.h"
#include "llvm/Support/Casting.h"

#define DEBUG_TYPE "optplayground"

#include "CustomPasses.h"
#include "InstanceInfo.h"

#include <cmath>
#include <map>
#include <set>
#include <vector>
#include <algorithm>
#include <string>
#include <sstream>

using namespace llvm;
using namespace llvm::PatternMatch;

namespace cellml {
namespace opt {

STATISTIC(NumHighOrderPowers, "number of high-order powers transformed to multiplications");
STATISTIC(NumExpandedConstants, "number of expanded constants");
STATISTIC(NumConstantsPropagatedInExp, "number of constants propagated into exp");
STATISTIC(NumConstantsFoldedIntoMul, "number of constants folded into multiplication next to exp");
STATISTIC(NumConstMultsReplacedByAdd_2m, "number of constant multiplications replaced by additions (*2)");
STATISTIC(NumConstMultsReplacedByAdd_1a, "number of constant multiplications replaced by additions (+1)");
STATISTIC(NumConstMultsReplacedByAdd_1s, "number of constant multiplications replaced by additions (-1)");

static Instruction *makeUnsafe(Instruction *I) {
	return I->setHasUnsafeAlgebra(true), I;
}

class FactorTable {
	Instruction                        *X = nullptr;
	std::map<ConstantFP*, Instruction*> map;
	std::vector<ConstantFP*>            posFactors, negFactors;

	Instruction *ensureAvailability(Instruction *Y, Instruction *Z, DominatorTree& DT) {
		if (! DT.dominates(Y, Z)) {
			Instruction *YY = Y->clone();
			YY->setName(Y->getName() + "ensav");

			BasicBlock *NCD = DT.findNearestCommonDominator(Y->getParent(), Z->getParent());
			assert (NCD);
			if (NCD == Z->getParent())
				YY->insertBefore(Z);
			else
				YY->insertBefore(NCD->getTerminator());

			Y->replaceAllUsesWith(YY);
			return YY;
		}

		return nullptr;
	}

public:
	void init(Instruction *X) {
		this->X = X;

		ConstantFP *One = cast<ConstantFP>(ConstantFP::get(X->getType(), 1.0));
		map.insert(std::make_pair(One, X));
		posFactors.push_back(One);
	}

	void add(BinaryOperator *Y) {
		assert (Y->getOpcode() == BinaryOperator::BinaryOps::FMul);
		assert (Y->getOperand(0) == X);

		ConstantFP *F = dyn_cast<ConstantFP>(Y->getOperand(1));
		assert (F);

		if (F->isZero())
			return;

		if (map.count(F))
			return; // will be CSE'd later. For now, just use the earliest occurence.

		map.insert(std::make_pair(F, Y));
		if (F->isNegative())
			negFactors.push_back(F);
		else
			posFactors.push_back(F);
	}

	void dump() {
		bool changed = false;

		std::sort(negFactors.begin(), negFactors.end(),
				[] (ConstantFP *A, ConstantFP *B) -> bool {
			return ! ConstantExpr::getFCmp(FCmpInst::Predicate::FCMP_OGT, A, B)->isZeroValue();
		}
		);
		std::sort(posFactors.begin(), posFactors.end(),
				[] (ConstantFP *A, ConstantFP *B) -> bool {
			return ! ConstantExpr::getFCmp(FCmpInst::Predicate::FCMP_OLT, A, B)->isZeroValue();
		}
		);

		std::vector<ConstantFP *> allFactors;
		allFactors.insert(allFactors.begin(), negFactors.begin(), negFactors.end());
		allFactors.insert(allFactors.begin(), posFactors.begin(), posFactors.end());

		errs() << "FactorTable for: " << *X << "\n";
		for (auto FI = allFactors.begin(), FE = allFactors.end(); FI != FE; FI++) {
			ConstantFP  *F   = *FI;
			ConstantFP  *Fm2 = cast<ConstantFP>(ConstantExpr::getFMul(F, ConstantFP::get(F->getType(), 2)));
			ConstantFP  *Fp1 = cast<ConstantFP>(ConstantExpr::getFAdd(F, ConstantFP::get(F->getType(), F->isNegative() ? -1 : 1)));
			Instruction *Y   = map.find(F)->second;

			errs() << "  " << format("%g", (*FI)->getValueAPF().convertToDouble()) << " -> " << *map.find(*FI)->second << "\n";
			if (map.count(Fm2)) {
				Instruction *Z = map.find(Fm2)->second;
				if (Z != X) { // prohibit 0.5*X -> X (infeasible)
					errs() << *Fm2 << " present (2x)\n";
				}
			}

			if (map.count(Fp1)) {
				Instruction *Z = map.find(Fp1)->second;
				errs() << *Fp1 << " present (+/- 1)\n";
			}
		}
	}

	bool optimise() {
		bool changed = false;

		std::sort(negFactors.begin(), negFactors.end(),
			[] (ConstantFP *A, ConstantFP *B) -> bool {
				return ! ConstantExpr::getFCmp(FCmpInst::Predicate::FCMP_OGT, A, B)->isZeroValue();
			}
		);
		std::sort(posFactors.begin(), posFactors.end(),
			[] (ConstantFP *A, ConstantFP *B) -> bool {
				return ! ConstantExpr::getFCmp(FCmpInst::Predicate::FCMP_OLT, A, B)->isZeroValue();
			}
		);

		std::vector<ConstantFP *> allFactors;
		allFactors.insert(allFactors.begin(), negFactors.begin(), negFactors.end());
		allFactors.insert(allFactors.begin(), posFactors.begin(), posFactors.end());

		std::set<ConstantFP *> replaced;
		replaced.insert(cast<ConstantFP>(ConstantFP::get(X->getType(), 1.0)));

		Function *P = X->getParent()->getParent();
		DominatorTree DT;
		DT.recalculate(*P);

		for (auto FI = allFactors.begin(), FE = allFactors.end(); FI != FE; FI++) {
			ConstantFP  *F   = *FI;
			ConstantFP  *Fm2 = cast<ConstantFP>(ConstantExpr::getFMul(F, ConstantFP::get(F->getType(), 2.0)));
			ConstantFP  *Fa1 = cast<ConstantFP>(ConstantExpr::getFAdd(F, ConstantFP::get(F->getType(), F->isNegative() ? -1.0 : 1.0)));
			Instruction *Y   = map.find(F)->second;

			if (map.count(Fm2) && ! replaced.count(Fm2)) {
				Instruction *Z = map.find(Fm2)->second;
				if (Instruction *YY = ensureAvailability(Y, Z, DT))
					map[F] = Y = YY;

				Instruction *NewAdd = makeUnsafe(BinaryOperator::CreateFAdd(Y, Y, "ftm", Z));
				Z->replaceAllUsesWith(NewAdd);

				map[Fm2] = Z;
				replaced.insert(Fm2);

				++NumConstMultsReplacedByAdd_2m;
				changed = true;
			}

			if (map.count(Fa1) && replaced.count(Fa1)) {
				Instruction *Z = map.find(Fa1)->second;
				if (Instruction *YY = ensureAvailability(Y, Z, DT))
					map[F] = Y = YY;

				Instruction *NewAddSub = makeUnsafe(F->isNegative() ?
					BinaryOperator::CreateFAdd(Y, X, "fta", Z) :
					BinaryOperator::CreateFSub(Y, X, "fts", Z)
				);
				Z->replaceAllUsesWith(NewAddSub);

				map[Fa1] = Z;
				replaced.insert(Fa1);

				if (! F->isNegative())
					++NumConstMultsReplacedByAdd_1a;
				else
					++NumConstMultsReplacedByAdd_1s;
				changed = true;
			}
		}

		return changed;
	}

};

class OptimisationPlayground : public FunctionPass {
	StringRef ModelName, OptimisationLevelName;

	Instruction *transformPowerOfIntegerConstant(Instruction *I) {
		Value      *X;
		ConstantFP *ExpFP;

		auto pattern = m_Intrinsic<Intrinsic::pow>(
				m_Value(X),
				m_ConstantFP(ExpFP));

		if (match(I, pattern) && I->hasUnsafeAlgebra()) {
			APSInt Exp(32, false); bool wasExact;
			if (APFloat::opOK == ExpFP->getValueAPF().convertToInteger(Exp, APFloat::rmTowardZero, &wasExact) && wasExact) {
				int N = Exp.getSExtValue();

				if (N > 2 && N <= 46) {
					/* From: Knuth, D. E. The Art of Computer Programming, Volume 2: Seminumerical Algorithms */
					/* A1 */
					Value *Y = ConstantFP::get(X->getType(), 1.0);
					Value *Z = X;

					for (;;) {
						/* A2 */
						bool odd = N & 1;
						N = N / 2;

						if (odd) {
							/* A3 */
							Y = makeUnsafe(BinaryOperator::CreateFMul(Z, Y, "knuthmul", I));

							/* A4 */
							if (N == 0)
								break;
						}

						/* A5 */
						Z = makeUnsafe(BinaryOperator::CreateFMul(Z, Z, "knuthsquare", I));
					}

					I->replaceAllUsesWith(Y);

					NumHighOrderPowers++;
					return cast<Instruction>(Y);
				}
			}
		}

		return nullptr;
	}

	Instruction *expandAlmostConstantExprs(Instruction *I) {
		Value *X;
		ConstantFP *C1, *C2;

		auto patternA = m_FMul(
				m_FAdd(
					m_Value(X),
					m_ConstantFP(C1)
				),
				m_ConstantFP(C2));

		auto patternS = m_FMul(
				m_FSub(
					m_ConstantFP(C1),
					m_Value(X)
				),
				m_ConstantFP(C2));

		bool isSub = false;
		if ((match(I, patternA) || (isSub = match(I, patternS))) && I->hasUnsafeAlgebra()) {
			ConstantFP *C12 = cast<ConstantFP>(ConstantExpr::getFMul(C1, C2));
			if (C12->getValueAPF().isNormal()) {
				Instruction *NewMul = makeUnsafe(BinaryOperator::CreateFMul(X, C2, "acemul", I));
				Instruction *NewAddSub = makeUnsafe(
						isSub ?
						BinaryOperator::CreateFSub(C12, NewMul, "acesub", I) :
						BinaryOperator::CreateFAdd(NewMul, C12, "aceadd", I));

				I->replaceAllUsesWith(NewAddSub);

				NumExpandedConstants++;
				return NewAddSub;
			}
		}

		return nullptr;
	}

	Instruction *optimiseExp(Instruction *I) {
		Value *X;
		ConstantFP *C1, *C2;

		auto patternFoldIntoMul = m_FMul(
				m_Intrinsic<Intrinsic::exp>(
					m_FAdd(
						m_Value(X),
						m_ConstantFP(C1)
					)
				),
				m_ConstantFP(C2));

		auto patternPropIntoExponent = m_FMul(
				m_Intrinsic<Intrinsic::exp>(
					m_Value(X)
				),
				m_ConstantFP(C2));

		if (match(I, patternFoldIntoMul) && I->hasUnsafeAlgebra()) {
			double expC1 = exp(C1->getValueAPF().convertToDouble());
			ConstantFP *ExpC1 = cast<ConstantFP>(ConstantFP::get(I->getType(), expC1));
			if (! ExpC1->getValueAPF().isNormal())
				return nullptr;

			Module *M = I->getParent()->getParent()->getParent();
			Instruction *NewExp = makeUnsafe(CallInst::Create(
					Intrinsic::getDeclaration(M, Intrinsic::exp, ArrayRef<Type *>(I->getType())),
					ArrayRef<Value *>(X), "exptomul", I));
			ConstantFP  *C3 = cast<ConstantFP>(ConstantExpr::getFMul(ExpC1, C2));
			Instruction *NewMul = makeUnsafe(BinaryOperator::CreateFMul(NewExp, C3, "exptomul", I));

			I->replaceAllUsesWith(NewMul);

			NumConstantsFoldedIntoMul++;
			return NewMul;
		}

		if (match(I, patternPropIntoExponent) && I->hasUnsafeAlgebra()) {
			if (I->getOperand(0)->getNumUses() > 1 || C2->isNegative() || C2->isZero())
				return nullptr;

			double logC2 = log(C2->getValueAPF().convertToDouble());
			ConstantFP *LogC2 = cast<ConstantFP>(ConstantFP::get(I->getType(), logC2));
			if (! LogC2->getValueAPF().isNormal())
				return nullptr;

			Module *M = I->getParent()->getParent()->getParent();
			Instruction *NewAdd = makeUnsafe(BinaryOperator::CreateFAdd(X, LogC2, "multoexp", I));
			Instruction *NewExp = makeUnsafe(CallInst::Create(
					Intrinsic::getDeclaration(M, Intrinsic::exp, ArrayRef<Type *>(I->getType())),
					ArrayRef<Value *>(NewAdd), "multoexp", I));

			I->replaceAllUsesWith(NewExp);

			NumConstantsPropagatedInExp++;
			return NewExp;
		}

		return nullptr;
	}

	bool reuseConstantFactors(Function &F) {
		std::map<Instruction *, FactorTable> Tables;

		for (auto II = inst_iterator(F), IE = inst_end(F); II != IE; II++) {
			Instruction *I = &*II;

			Value      *V;
			ConstantFP *C;
			auto pattern = m_FMul(m_Value(V), m_ConstantFP(C));

			if (match(I, pattern)) {
				Instruction *X = dyn_cast<Instruction>(V);
				if (! X || ! X->hasUnsafeAlgebra())
					continue;

				if (Tables.count(X) == 0)
					Tables[X].init(X);

				Tables[X].add(cast<BinaryOperator>(I));
			}
		}

		bool changed = false;
		for (auto TI = Tables.begin(), TE = Tables.end(); TI != TE; TI++)
			changed |= TI->second.optimise();

		return changed;
	}

public:
	static char ID;
	OptimisationPlayground() : FunctionPass(ID) {}

	bool runOnFunction(Function &F) {
		bool changed = false;

		InstanceInfo &II = getAnalysis<InstanceInfo>();
		if (ModelName.empty())
			ModelName = II.ModelName;
		else
			assert (ModelName.equals(II.ModelName) && "Inconsistent statistics");

		if (OptimisationLevelName.empty())
			OptimisationLevelName = II.OptimisationLevelName;
		else
			assert (OptimisationLevelName.equals(II.OptimisationLevelName) && "Inconsistent statistics");

		for (auto II = inst_iterator(F); II != inst_end(F); ) {
			Instruction *I = &*(II++), *IT;

			if ((IT = transformPowerOfIntegerConstant(I))) { I = IT; changed = true; }
			if ((IT = expandAlmostConstantExprs(I)))       { I = IT; changed = true; }
			if ((IT = optimiseExp(I)))                     { I = IT; changed = true; }
		}

		changed |= reuseConstantFactors(F);

		return changed;
	}

	bool doFinalization(Module &M) {
		if (ModelName.empty() || OptimisationLevelName.empty())
			return false;

		std::ostringstream filename;
		filename << "optplygrnd." << OptimisationLevelName.str() << ".csv";
		std::string error;
		raw_fd_ostream StatFile(filename.str().c_str(), error, sys::fs::F_None);

		StatFile << ModelName                     << ',';
		StatFile << OptimisationLevelName         << ',';
		StatFile << NumHighOrderPowers            << ',';
		StatFile << NumExpandedConstants          << ',';
		StatFile << NumConstantsPropagatedInExp   << ',';
		StatFile << NumConstantsFoldedIntoMul     << ',';
		StatFile << NumConstMultsReplacedByAdd_2m << ',';
		StatFile << NumConstMultsReplacedByAdd_1a << ',';
		StatFile << NumConstMultsReplacedByAdd_1s << '\n';

		StatFile.close();

		return false;
	}

	void getAnalysisUsage(AnalysisUsage &AU) const {
		AU.addRequired<InstanceInfo>();
		AU.setPreservesCFG();
	}
};

char OptimisationPlayground::ID = 0;

Pass *createOptimisationPlaygroundPass() {
	return new OptimisationPlayground();
}

}
}
