/*******************************************************************************
 * Copyright 2015 Julian Oppermann, TU Darmstadt
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

#include "llvm/Pass.h"
#include "llvm/IR/Function.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/ADT/Statistic.h"

#include "Equations.h"
#include "EquationSystem.h"
#include "config.h"

#define DEBUG_TYPE "reuseexpressions"

using namespace llvm;

namespace cellml {
namespace opt {

class ReuseExpressions : public FunctionPass, ExprVisitor {
	ExprSet  reusableExprsS;
	ExprList reusableExprsL;

	unsigned tmpUid = 0;

	void findReusableExpr(Expr *expr) {
		if (expr->getSubstExpr())
			return;

		unsigned numUses = expr->getNumUses();
		assert (numUses > 0);
#ifndef CELLML_TO_VHDL_COMPAT_MODE
		if (numUses == 1)
			return;
#else
		if (numUses == 1 && expr->getKind() != ExprKind::Select)
			return; /* FIXME: always reuse SelectExprs, in order to make them a "toplevel RHS" */
		if (expr->getKind() == ExprKind::Binary && ((BinaryExpr *) expr)->getFunction() >= BinaryFunction::lt)
			return; /* FIXME: don't reuse comparisons */
#endif

		if (reusableExprsS.count(expr))
			return;

		reusableExprsS.insert(expr);
		reusableExprsL.push_back(expr);
	}

#ifdef CELLML_TO_VHDL_COMPAT_MODE
	bool isRatesRef(Expr *expr) {
		if (expr->getKind() != ExprKind::Var)
			return false;
		VarExpr *var = (VarExpr *) expr;
		return StringRef("RATES").equals(var->getBase()->getName());
	}
#endif

public:
	static char ID;
	ReuseExpressions() : FunctionPass(ID) {}

	bool runOnFunction(Function &F) {
		EquationSystem &equationSystem = getAnalysis<EquationSystem>();

		for (auto ESI = equationSystem.begin(); ESI != equationSystem.end(); ) {
			auto      ESIn  = ESI++;
			Equation *eq    = *ESIn;
			Expr     *eqLhs = eq->getLhs();
			Expr     *eqRhs = eq->getRhs();

			reusableExprsS.clear();
			reusableExprsL.clear();
			eq->getRhs()->accept(this);

#ifdef CELLML_TO_VHDL_COMPAT_MODE
			/* XXX: Workaround, do not reuse complete RHS of a RATES equation,
			 *      in order to keep RATES references out of other equations */
			bool forceTempForRatesRef = isRatesRef(eqLhs) && eqRhs->getNumUses() > 1;
#endif

			for (Expr *reusableExpr : reusableExprsL) {
#ifndef CELLML_TO_VHDL_COMPAT_MODE
				if (reusableExpr == eqRhs)
#else
				if (reusableExpr == eqRhs && !forceTempForRatesRef)
#endif
					continue; /* don't create a temp for the whole equation expression - see below */
				if (reusableExpr->getSubstExpr())
					continue; /* e.g., the expr was used in a previous equation */

				TmpVarExpr *tmpVar = new TmpVarExpr(tmpUid++);
				equationSystem.addExpr(tmpVar);
				equationSystem.addEquation(tmpVar, reusableExpr, ESIn);
				reusableExpr->setSubstExpr(tmpVar);
			}

#ifndef CELLML_TO_VHDL_COMPAT_MODE
			if (reusableExprsS.count(eqRhs))
#else
			if (reusableExprsS.count(eqRhs) && ! forceTempForRatesRef)
#endif
				eqRhs->setSubstExpr(eqLhs);
		}

		return false;
	}

	void getAnalysisUsage(AnalysisUsage &AU) const {
		AU.addRequired<EquationSystem>();
		AU.setPreservesAll();
	}

	void visit(VarExpr *var) { }
	void visit(TmpVarExpr *tmp) { }
	void visit(ConstExpr *konst) { }

	void visit(UnaryExpr *unary) {
		unary->getOp()->accept(this);
		findReusableExpr(unary);
	}

	void visit(BinaryExpr *binary) {
		binary->getLhs()->accept(this);
		binary->getRhs()->accept(this);
		findReusableExpr(binary);
	}

	void visit(SelectExpr *select) {
		select->getCond()->accept(this);
		select->getIfTrue()->accept(this);
		select->getIfFalse()->accept(this);
		findReusableExpr(select);
	}

};

char ReuseExpressions::ID = 0;

Pass *createExpressionReuserPass() {
	return new ReuseExpressions();
}

}
}
