/*******************************************************************************
 * Copyright 2015 Julian Oppermann, TU Darmstadt
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

#include "llvm/Pass.h"
#include "llvm/IR/Function.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Support/Format.h"

#include "Equations.h"
#include "EquationSystem.h"
#include "CustomPasses.h"
#include "config.h"

using namespace llvm;

namespace cellml {
namespace opt {

class SynthesiseC : public FunctionPass, public ExprVisitor {
	ExprSet emitted;

	/* returns true if a substitution is available and has already been emitted.
	 * otherwise, return false to indicate that given expression should be emitted normally. */
	bool handleSubstExpr(Expr *expr) {
		if (Expr *subst = expr->getSubstExpr()) {
			if (emitted.count(expr)) {
				subst->accept(this);
				return true;
			}
			emitted.insert(expr);
		}
		return false;
	}

public:
	static char ID;
	SynthesiseC() : FunctionPass(ID) {}

	void visitEquation(Equation *equation) {
		equation->getLhs()->accept(this);
		outs() << " = ";
		equation->getRhs()->accept(this);
		outs() << ";\n";
	}

	bool runOnFunction(Function &F) {
		emitted.clear();

		EquationSystem &equations = getAnalysis<EquationSystem>();

		outs()	<< "void\n"
				<< "computeRates(double VOI, double* CONSTANTS, double* RATES, double* STATES, double* ALGEBRAIC)\n"
				<< "{\n";

		if (unsigned nTmpVars = equations.getNTmpVars()) {
			outs() << "double TEMP[" << nTmpVars << "];\n";
		}

		for (Equation *eq : equations) {
			visitEquation(eq);
		}

		outs() << "}\n";

		return false;
	}

	void getAnalysisUsage(AnalysisUsage &AU) const {
		AU.addRequired<EquationSystem>();
		AU.setPreservesAll();
	}

	void visit(VarExpr *var) {
		outs() << var->getBase()->getName();
		if (var->getIndex())
			outs() << "[" << var->getIndex()->getUniqueInteger() << "]";
	}

	void visit(TmpVarExpr *tmp) {
		outs() << "TEMP[" << tmp->getTmpId() << "]";
	}

	void visit(ConstExpr *konst) {
#ifndef CELLML_TO_VHDL_COMPAT_MODE
		outs() << "(" << format("%.9E", konst->getFloatValue()->getValueAPF().convertToDouble()) << ")";
#else
		outs() << "(" << format("%.9e", konst->getFloatValue()->getValueAPF().convertToDouble()) << ")";
#endif
	}

	void visit(UnaryExpr *unary) {
		assert (unary->getOp());

		if (handleSubstExpr(unary))
			return;

		outs() << "(";

		switch (unary->getFunction()) {
		case UnaryFunction::exp:   outs() << "exp";   break;
		case UnaryFunction::log:   outs() << "log";   break;
		case UnaryFunction::floor: outs() << "floor"; break;
		case UnaryFunction::abs:   outs() << "fabs";  break;
		case UnaryFunction::neg:   outs() << "-";     break;
		default:
			assert (false && "unhandled unary function");
		}

		outs() << "(";
		unary->getOp()->accept(this);
		outs() << "))";
	}

	void visit(BinaryExpr *binary) {
		assert (binary->getLhs() && binary->getRhs());

		if (handleSubstExpr(binary))
			return;

		BinaryFunction func = binary->getFunction();
		if (func == BinaryFunction::pow) {
			outs() << "(pow(";
			binary->getLhs()->accept(this);
			outs() << ", ";
			binary->getRhs()->accept(this);
			outs() << "))";
			return;
		}

		outs() << "(";
		binary->getLhs()->accept(this);

		switch (func) {
		case BinaryFunction::add:  outs() << " + ";   break;
		case BinaryFunction::sub:  outs() << " - ";   break;
		case BinaryFunction::mul:  outs() << " * ";   break;
		case BinaryFunction::div:  outs() << " / ";   break;
		case BinaryFunction::lt:   outs() << " < ";   break;
		case BinaryFunction::le:   outs() << " <= ";  break;
		case BinaryFunction::gt:   outs() << " > ";   break;
		case BinaryFunction::ge:   outs() << " >= ";  break;
		case BinaryFunction::land: outs() << " && ";  break;
#ifndef CELLML_TO_VHDL_COMPAT_MODE
		case BinaryFunction::eq:   outs() << " == ";  break;
		case BinaryFunction::ne:   outs() << " != ";  break;
		case BinaryFunction::lor:  outs() << " || ";  break;
#endif
		default:
			assert (false && "unhandled binary function");
		}

		binary->getRhs()->accept(this);
		outs() << ")";
	}

	void visit(SelectExpr *select) {
		assert (select->getCond() && select->getIfTrue() && select->getIfFalse());

		if (handleSubstExpr(select))
			return;

#ifndef CELLML_TO_VHDL_COMPAT_MODE
		outs() << "(";
#endif
		select->getCond()->accept(this);
		outs() << " ? ";
		select->getIfTrue()->accept(this);
		outs() << " : ";
		select->getIfFalse()->accept(this);
#ifndef CELLML_TO_VHDL_COMPAT_MODE
		outs() << ")";
#endif
	}
};

char SynthesiseC::ID = 0;

Pass *createCSynthesiserPass() {
	return new SynthesiseC();
}

}
}
