/*******************************************************************************
 * Copyright 2015 Julian Oppermann, TU Darmstadt
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

#include "llvm/Pass.h"
#include "llvm/IR/Function.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Support/FileSystem.h"
#include "llvm/Support/Format.h"

#include "Equations.h"
#include "EquationSystem.h"
#include "InstanceInfo.h"

#include "CustomPasses.h"

#include <string>
#include <sstream>

using namespace llvm;

namespace cellml {
namespace opt {

class WriteDot : public FunctionPass, public ExprVisitor {
	raw_fd_ostream *DotFile = nullptr;
	ExprSet         emitted;

	bool handleEmitted(Expr *expr) {
		if (emitted.count(expr))
			return true;
		emitted.insert(expr);
		return false;
	}

	void emitNodeStart(Expr *expr) {
		*DotFile << "n" << expr->getId() << " [label=\"";
	}

	void emitNodeStart(Equation *equation) {
		*DotFile << "ne" << equation->getId() << " [label=\"";
	}

	void emitNodeEnd() {
		*DotFile << "\"];\n";
	}

	void emitEdge(Expr *src, Expr *dst) {
		*DotFile << "n" << src->getId() << " -> n" << dst->getId() << ";\n";
	}

	void emitEdge(Expr *src, Equation *dst) {
		*DotFile << "n" << src->getId() << " -> ne" << dst->getId() << ";\n";
	}

public:
	static char ID;
	WriteDot() : FunctionPass(ID) {}

	void visitEquation(Equation *equation, Function &F) {
		if (equation->getLhs()->getKind() == ExprKind::TmpVar)
			return;

		emitted.clear();

		InstanceInfo &II = getAnalysis<InstanceInfo>();

		std::ostringstream filename;
		filename << "eqs." << II.OptimisationLevelName.str() << "." << equation->getId() << ".dot";
		std::string error;
		DotFile = new raw_fd_ostream(filename.str().c_str(), error, sys::fs::F_None);

		*DotFile << "digraph G {\n";
		equation->getLhs()->accept(this);
		equation->getRhs()->accept(this);

		emitNodeStart(equation);
		*DotFile << "=";
		emitNodeEnd();

		emitEdge(equation->getLhs(), equation);
		emitEdge(equation->getRhs(), equation);

		*DotFile << "}\n";

		DotFile->close();
		delete DotFile;
		DotFile = nullptr;
	}

	bool runOnFunction(Function &F) {
		EquationSystem &equations = getAnalysis<EquationSystem>();

		for (Equation *eq : equations) {
			visitEquation(eq, F);
		}

		return false;
	}

	void getAnalysisUsage(AnalysisUsage &AU) const {
		AU.addRequired<EquationSystem>();
		AU.addRequired<InstanceInfo>();
		AU.setPreservesAll();
	}

	void visit(VarExpr *var) {
		if (handleEmitted(var))
			return;

		emitNodeStart(var);
		*DotFile << var->getBase()->getName();
		if (var->getIndex())
			*DotFile << "[" << var->getIndex()->getUniqueInteger() << "]";
		emitNodeEnd();
	}

	void visit(TmpVarExpr *tmp) {
		if (handleEmitted(tmp))
			return;

		emitNodeStart(tmp);
		*DotFile << "tmp" << tmp->getTmpId();
		emitNodeEnd();
	}

	void visit(ConstExpr *konst) {
		if (handleEmitted(konst))
			return;

		emitNodeStart(konst);
		*DotFile << format("%g", konst->getFloatValue()->getValueAPF().convertToDouble());
		emitNodeEnd();
	}

	void visit(UnaryExpr *unary) {
		assert (unary->getOp());

		if (handleEmitted(unary))
			return;

		unary->getOp()->accept(this);

		emitNodeStart(unary);

		switch (unary->getFunction()) {
		case UnaryFunction::exp:   *DotFile << "exp";   break;
		case UnaryFunction::log:   *DotFile << "log";   break;
		case UnaryFunction::floor: *DotFile << "floor"; break;
		case UnaryFunction::abs:   *DotFile << "fabs";  break;
		case UnaryFunction::neg:   *DotFile << "neg";   break;
		}

		emitNodeEnd();

		emitEdge(unary->getOp(), unary);
	}

	void visit(BinaryExpr *binary) {
		assert (binary->getLhs() && binary->getRhs());

		if (handleEmitted(binary))
			return;

		binary->getLhs()->accept(this);
		binary->getRhs()->accept(this);

		emitNodeStart(binary);

		switch (binary->getFunction()) {
		case BinaryFunction::add:  *DotFile << "+";   break;
		case BinaryFunction::sub:  *DotFile << "-";   break;
		case BinaryFunction::mul:  *DotFile << "*";   break;
		case BinaryFunction::div:  *DotFile << "/";   break;
		case BinaryFunction::pow:  *DotFile << "pow"; break;
		case BinaryFunction::lt:   *DotFile << "<";   break;
		case BinaryFunction::le:   *DotFile << "<=";  break;
		case BinaryFunction::gt:   *DotFile << ">";   break;
		case BinaryFunction::ge:   *DotFile << ">=";  break;
		case BinaryFunction::land: *DotFile << "&&";  break;
		default:
			assert (false && "unhandled binary function");
		}

		emitNodeEnd();

		emitEdge(binary->getLhs(), binary);
		emitEdge(binary->getRhs(), binary);
	}

	void visit(SelectExpr *select) {
		assert (select->getCond() && select->getIfTrue() && select->getIfFalse());

		if (handleEmitted(select))
			return;

		select->getCond()->accept(this);
		select->getIfTrue()->accept(this);
		select->getIfFalse()->accept(this);

		emitNodeStart(select);
		*DotFile << "select";
		emitNodeEnd();

		emitEdge(select->getCond(), select);
		emitEdge(select->getIfTrue(), select);
		emitEdge(select->getIfFalse(), select);
	}
};

char WriteDot::ID = 0;

Pass *createDotWriterPass() {
	return new WriteDot();
}

}
}
