################################################################################
# Copyright 2015 Julian Oppermann, TU Darmstadt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
#
# cellml-opt.py -- Driver script for the LLVM-based optimisation
#                  of CellML models.
#
################################################################################

import sys
import optparse
import string
import re
import os
import subprocess
import tempfile

from config import CLANG, OPT, CELLML_OPT

def parseFile(file):
	f = open(file, "r")
	s = f.read()
	
	begin = 0
	end = 0
	
	begin = s.index("/*", end)
	end   = s.index("*/", begin)+len("*/\n")
	varSpec = s[begin:end]

	begin = s.index("/*", end)
	end   = s.index("*/", begin)+len("*/\n")
	description = s[begin:end]

	begin = s.index("void\ninitConsts", end)
	end   = s.index("}", begin)+len("}\n")
	initConsts = s[begin:end]
	
	begin = s.index("void\ncomputeRates", end)
	end   = s.index("}", begin)+len("}\n")
	computeRates = s[begin:end]
	
	begin = s.index("void\ncomputeVariables", end)
	end   = s.index("}", begin)+len("}\n")
	computeVariables = s[begin:end]
	
	f.close()
	
	return (varSpec, description, initConsts, computeRates, computeVariables)
	
def runCmd(cmd, stdin):
	p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
	(stdout, stderr) = p.communicate(stdin)
	stderr = stderr.strip()
	if verbose and len(stderr) > 0:
		print stderr
	if p.returncode != 0:
		raise Exception("invocation of '%s' failed with err=%d\n" % (cmd[0], p.returncode))
	return stdout

def compileToLLVMIR(computeRates, optLevel):
	computeRates = "#include <math.h>\n" + makeEager(computeRates)
	cmd = [CLANG, "-x", "c", "-fmath-errno"] # setting -fmath-errno to force (frontend) usage of pow() instead of llvm.pow() intrinsic
	if optLevel in ["zU", "EqC", "zEqC"]:
		cmd += ["-ffast-math"]
	cmd += ["-S", "-emit-llvm", "-o", "-", "-"] 
	return runCmd(cmd, computeRates)
	
def optimise(cellMLName, LLVMIR, optLevel):
	# for -Oz[U], do a separate 'opt' run, and then synthesise through cellml-opt's -O0
	if optLevel in ["z", "zU", "zEqC"]:
		cmd = [OPT, "-S", "-Oz"]
		if optLevel in ["zU", "zEqC"]:
			cmd += ["-enable-unsafe-fp-math"]
		LLVMIR = runCmd(cmd, LLVMIR)
		
	cmd  = [CELLML_OPT, "-O" + optLevel, "-model-name=" + cellMLName]
	cmd += ["-enable-pre=false", "-enable-load-pre=false"]
	if optLevel in ["EqC", "zEqC"]:
		cmd += ["-enable-unsafe-fp-math"]
	
	if verbose:
		cmd += ["-stats"]
		#if optLevel in ["0", "EqR", "EqC"]:
		#	cmd += ["-dot-eqs"]                    # don't spam the directory with the -Oz[U] equations
	
	if debug:
		tempFile = tempfile.NamedTemporaryFile(suffix=".ll", dir="./", delete=False)
		tempFile.write(LLVMIR);
		tempFile.close();
		os.execvp("lldb", ["lldb"] + [CELLML_OPT, "--"] + cmd[1:] + [tempFile.name])
	
	return runCmd(cmd, LLVMIR)
	
def writeOutput(file, varSpec, description, initConsts, computeRates, computeVariables):
	f = open(file, "w")
	f.write(varSpec)
	f.write(description)
	f.write(initConsts)
	f.write(computeRates)
	f.write(computeVariables)
	f.close()

def convertToFloat(s):
	s = string.replace(s, "double", "float")
	s = string.replace(s, "pow",    "powf")
	s = string.replace(s, "exp",    "expf")
	s = string.replace(s, "log",    "logf")
	s = string.replace(s, "floor",  "floorf")
	s = string.replace(s, "fabs",   "fabsf")	
	s = re.sub(r"([-+]?((\d+\.(\d*)|\.\d+)([eE][-+]?\d+)?)|(\d+[eE][-+]?\d+))", r"\1f", s)
	return s

def makeEager(s):
	return string.replace(s, "&&", "&")

def main():
	usage = "usage: %prog [options] <CellML model>"
	parser = optparse.OptionParser(usage)
	parser.add_option("-o", dest="outfile", help="specify output filename")
	parser.add_option("-O", dest="optLevel", 
			help="""specify what optimizations to run in cellml-opt
			Possible values:
			'Ref'    Source code pass-through
			'0'      cellml-opt:     No optimizations
			'EqR'    cellml-opt:     Enable tailored recipe of built-in FP-safe optimisations
			'EqC'    cellml-opt:     Enable tailored unsafe recipe of built-in optimisations and custom transformations
			'z'      opt:            Use preset -Oz
			'zU'     opt:            Use preset -Oz, and enable unsafe math
			'zEqC'   opt+cellml-opt: like OEqC, but assume -Oz -ffast-math outside of cellml-opt""")
	parser.add_option("--debug", dest="debug", action="store_true", help="show debug cellml-opt command")
	parser.add_option("-v", "--verbose", dest="verbose", action="store_true", help="enable verbose output")
	parser.add_option("-F", "--floatify", dest="floatify", action="store_true", help="convert the output to use single-precision FP")
	parser.add_option("-S", "--synth", dest="synth", action="store_true", help="don't emit math.h include")
	(options, args) = parser.parse_args()
	
	if len(args) != 1:
		parser.error("no CellML input file specified")
		
	cellMLFile = args[0]
	(cellMLName, cellMLExt) = os.path.splitext(cellMLFile)
	if cellMLExt != ".c":
		parser.error("CellML input '%s' is not a C file" % cellMLFile)
	
	global debug, verbose
	debug   = options.debug
	verbose = options.verbose
	
	optLevel = options.optLevel or "0"
	
	(varSpec, description, initConsts, computeRates, computeVariables) = parseFile(cellMLFile)
	
	if optLevel != "Ref":
		LLVMIR = compileToLLVMIR(computeRates, optLevel)
		computeRatesOpt = optimise(cellMLName, LLVMIR, optLevel)
	else:
		computeRatesOpt = makeEager(computeRates)
	
	m = re.search(r"double TEMP\[(\d+)\]", computeRatesOpt)
	if m:
		nTmpVars = int(m.group(1))
		varSpecLines = varSpec.splitlines()
		varSpecLines = varSpecLines[:-1] + ["   There are a total of %d entries in the temp variable array." % nTmpVars] + varSpecLines[-1:]
		varSpec = "\n".join(varSpecLines) + "\n"
	
	if not options.synth:
		initConsts = "#include <math.h>\n" + initConsts
	
	if options.floatify:
		initConsts       = convertToFloat(initConsts)
		computeRatesOpt  = convertToFloat(computeRatesOpt)
		computeVariables = convertToFloat(computeVariables)
	
	writeOutput(options.outfile or ("%s.%s.c" % (cellMLName, optLevel)), varSpec, description, initConsts, computeRatesOpt, computeVariables)

if __name__ == "__main__":
	main()
