################################################################################
# Copyright 2015 Julian Oppermann, TU Darmstadt
#                Oliver Sinnen, University of Auckland
#                Andreas Koch, TU Darmstadt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
#
# evaluation.py -- compute metrics for the FPL'2015 paper.
#
################################################################################

import csv
import optparse
import math

class OptResult:
	def __init__(self, row):
		assert len(row) == 19
		
		(
		self.level,
		self.nAdd,
		self.nSub,
		self.nMul,
		self.nDiv,
		self.nPow,
		self.nExp,
		self.nLog,
		self.nFlr,
		self.nAbs,
		self.nNeg,
		self.nCmp,
		self.nBit,
		self.nLUT,
		self.nDLR,
		self.nALM,
		self.nDSP,
		self.errAbs,
		self.errRel
		) = tuple(row[0:1] + [int(i) for i in row[1:-2]] + [float(f) for f in row[-2:]])
		
		self.nAddSub = self.nAdd + self.nSub
		self.nOther  = self.nLog + self.nFlr + self.nAbs + self.nNeg + self.nCmp + self.nBit
		self.nTotal  = reduce(lambda a, b: a + b, [int(i) for i in row[1:-6]])
		
		self.cost    = math.sqrt(ofALMs(self.nALM)**2 + ofDSPs(self.nDSP)**2)
		
class OptApplications:
	def __init__(self, row):
		assert len(row) == 7
		
		(
		self.nHighOrderPowers,
		self.nExpandedConstants,
		self.nConstantsPropagatedInExp,
		self.nConstantsFoldedIntoMul,
		self.nConstMultsReplacedByAdd_2m,
		self.nConstMultsReplacedByAdd_1a,
		self.nConstMultsReplacedByAdd_1s
		) = tuple([int(i) for i in row])
		
		self.nConstMultsReplacedByAdd = self.nConstMultsReplacedByAdd_2m + self.nConstMultsReplacedByAdd_1a + self.nConstMultsReplacedByAdd_1s

class ModelResult:
	def __init__(self, modelName):
		self.name            = modelName
		self.optResults      = {"0": None, "z": None, "zU": None, "EqC": None, "zEqC": None}
		self.optApplications = {"EqC": None, "zEqC": None}
		
	def hasMinCost(self, optLevel):
		assert optLevel in self.optResults
		sortedByCost = sorted(list(self.optResults.values()), cmp=lambda a, b: cmp(a.cost, b.cost))
		return (sortedByCost[0].level == optLevel)

def ofAll(n):    return float(n)/float(nModels)*100.0
def ofALMs(n):   return float(n)/212480.0*100.0
def ofDSPs(n):   return float(n)/1024.0*100.0
def norm(x, to): return float(x)/float(to)
	
def sum(a, b):   return a + b
def sumT2(a, b): return (a[0] + b[0], a[1] + b[1])
def sumT3(a, b): return (a[0] + b[0], a[1] + b[1], a[2] + b[2])

def dumpModelNames(modelNames, fileName):
	if not modelNames:
		return
	
	with open(fileName, 'w') as f:
		for m in modelNames:
			f.write(m + '\n')

optLevels = ["0", "z", "zU", "EqC", "zEqC"]

usage = "usage: %prog [options] <stats csv file> <optplygrnd csv file>"
parser = optparse.OptionParser(usage)
parser.add_option("-T", "--threshold", dest="threshold")
parser.add_option("--detailed", action="store_true", dest="detailed")
(options, args) = parser.parse_args()

if (len(args) != 2):
	parser.error("Please specify two csv files")
	
models = dict()

incompleteModels = set()
with open(args[0], 'r') as csvfile:
	statsreader = csv.reader(csvfile, delimiter=',')
	for row in statsreader:
		if (len(row) != 20):
			incompleteModels.add(row[0])
			continue # probably missing one error elements. Just ignore those
		
		modelName = row[0]
		if modelName not in models:
			models[modelName] = ModelResult(modelName)

		optName = row[1]
		models[modelName].optResults[optName] = OptResult(row[1:])

with open(args[1], 'r') as csvfile:
	statsreader = csv.reader(csvfile, delimiter=',')
	for row in statsreader:
		if (len(row) != 9):
			incompleteModels.add(row[0])
			continue # something weird
		
		modelName = row[0]
		optName   = row[1]
		if modelName not in models:
			continue
		
		assert optName in ["EqC", "zEqC"]
				
		models[modelName].optApplications[optName] = OptApplications(row[2:])

dumpModelNames(incompleteModels, "incompleteModels.txt")

###############################################################################

unsuitableModels = [m.name for m in models.values() if m.optResults["z"].errRel > 0.0]
for model in unsuitableModels:
	del models[model]

dumpModelNames(unsuitableModels, "unsuitableModels.txt")

###############################################################################

nModels = len(models)
print "Total number of models: %d\n" % nModels

###############################################################################

threshold = float(options.threshold or 1.0)

accurateModels  = [m.name for m in models.values() if m.optResults["zU"].errRel <= threshold and m.optResults["EqC"].errRel <= threshold and m.optResults["zEqC"].errRel <= threshold]
nAccurateModels = len(accurateModels)
average_zU      = reduce(sum, [models[mn].optResults["zU"].errRel for mn in accurateModels]) / float(nAccurateModels)
average_EqC     = reduce(sum, [models[mn].optResults["EqC"].errRel for mn in accurateModels]) / float(nAccurateModels)
average_zEqC    = reduce(sum, [models[mn].optResults["zEqC"].errRel for mn in accurateModels]) / float(nAccurateModels)
print """Models below %.3f %% relative error: %d (%.2f%%)
  avg. zU:    %.3f
  avg. EqC:   %.3f
  avg. zEqC:  %.3f
""" % (threshold, nAccurateModels, ofAll(nAccurateModels), average_zU, average_EqC, average_zEqC)

for m in models.values():
	for ol in optLevels[2:]:
		if m.optResults[ol].errRel > threshold:
			m.optResults[ol] = m.optResults["z"]

###############################################################################

nModelsMinCost_0    = len([m.name for m in models.values() if m.hasMinCost("0")])
nModelsMinCost_z    = len([m.name for m in models.values() if m.hasMinCost("z")])
nModelsMinCost_zU   = len([m.name for m in models.values() if m.hasMinCost("zU")])
nModelsMinCost_EqC  = len([m.name for m in models.values() if m.hasMinCost("EqC")])
nModelsMinCost_zEqC = len([m.name for m in models.values() if m.hasMinCost("zEqC")])

print """Min cost with:
  0:     %d (%.2f%%)
  z:     %d (%.2f%%)
  zU:    %d (%.2f%%)
  EqC:   %d (%.2f%%)
  zEqC:  %d (%.2f%%)
""" % (
nModelsMinCost_0,    ofAll(nModelsMinCost_0),
nModelsMinCost_z,    ofAll(nModelsMinCost_z),
nModelsMinCost_zU,   ofAll(nModelsMinCost_zU),
nModelsMinCost_EqC,  ofAll(nModelsMinCost_EqC),
nModelsMinCost_zEqC, ofAll(nModelsMinCost_zEqC)
)

###############################################################################

nOverall_0    = reduce(sumT3, [(m.optResults["0"].nALM, m.optResults["0"].nDSP, m.optResults["0"].cost) for m in models.values()])
nOverall_z    = reduce(sumT3, [(m.optResults["z"].nALM, m.optResults["z"].nDSP, m.optResults["z"].cost) for m in models.values()])
nOverall_zU   = reduce(sumT3, [(m.optResults["zU"].nALM, m.optResults["zU"].nDSP, m.optResults["zU"].cost) for m in models.values()])
nOverall_EqC  = reduce(sumT3, [(m.optResults["EqC"].nALM, m.optResults["EqC"].nDSP, m.optResults["EqC"].cost) for m in models.values()])
nOverall_zEqC = reduce(sumT3, [(m.optResults["zEqC"].nALM, m.optResults["zEqC"].nDSP, m.optResults["zEqC"].cost) for m in models.values()])


print """Overall resource usage:
          ALM   DSP   Cost
  0:     %.3f %.3f %.3f
  z:     %.3f %.3f %.3f
  zU:    %.3f %.3f %.3f
  EqC:   %.3f %.3f %.3f
  zEqC:  %.3f %.3f %.3f
""" % (
norm(nOverall_0[0], nOverall_0[0]), norm(nOverall_0[1], nOverall_0[1]), norm(nOverall_0[2], nOverall_0[2]),
norm(nOverall_z[0], nOverall_0[0]), norm(nOverall_z[1], nOverall_0[1]), norm(nOverall_z[2], nOverall_0[2]),
norm(nOverall_zU[0], nOverall_0[0]), norm(nOverall_zU[1], nOverall_0[1]), norm(nOverall_zU[2], nOverall_0[2]),
norm(nOverall_EqC[0], nOverall_0[0]), norm(nOverall_EqC[1], nOverall_0[1]), norm(nOverall_EqC[2], nOverall_0[2]),
norm(nOverall_zEqC[0], nOverall_0[0]), norm(nOverall_zEqC[1], nOverall_0[1]), norm(nOverall_zEqC[2], nOverall_0[2])
)

###############################################################################

nHighOrderPowers          = reduce(sumT2, [(m.optApplications["EqC"].nHighOrderPowers, m.optApplications["zEqC"].nHighOrderPowers) for m in models.values()])
nExpandedConstants        = reduce(sumT2, [(m.optApplications["EqC"].nExpandedConstants, m.optApplications["zEqC"].nExpandedConstants) for m in models.values()])
nConstantsPropagatedInExp = reduce(sumT2, [(m.optApplications["EqC"].nConstantsPropagatedInExp, m.optApplications["zEqC"].nConstantsPropagatedInExp) for m in models.values()])
nConstantsFoldedIntoMul   = reduce(sumT2, [(m.optApplications["EqC"].nConstantsFoldedIntoMul, m.optApplications["zEqC"].nConstantsFoldedIntoMul) for m in models.values()])
nConstMultsReplacedByAdd  = reduce(sumT2, [(m.optApplications["EqC"].nConstMultsReplacedByAdd, m.optApplications["zEqC"].nConstMultsReplacedByAdd) for m in models.values()])

print """Overall custom optimisation applications:
                         EqC   zEqC
  High-order powers:     %4d  %4d
  Constant expansion:    %4d  %4d
  Const prop. into exp:  %4d  %4d
  Const folded into mul: %4d  %4d
  MCM:                   %4d  %4d
""" % (nHighOrderPowers + nExpandedConstants + nConstantsPropagatedInExp + nConstantsFoldedIntoMul + nConstMultsReplacedByAdd)

###############################################################################

if options.detailed:
	dmodels = ["example1", "example2", "example3", "example4"]
	optLevels = optLevels[:-1]
	
	addSubL = []
	mulL    = []
	divL    = []
	expL    = []
	powL    = []
	otherL  = []
	totalL  = []
	for m in [models[m] for m in dmodels]:
		addSubL += [m.optResults[ol].nAddSub for ol in optLevels]
		mulL    += [m.optResults[ol].nMul for ol in optLevels]
		divL    += [m.optResults[ol].nDiv for ol in optLevels]
		expL    += [m.optResults[ol].nExp for ol in optLevels]
		powL    += [m.optResults[ol].nPow for ol in optLevels]
		otherL  += [m.optResults[ol].nOther for ol in optLevels]
		totalL  += [m.optResults[ol].nTotal for ol in optLevels]
		
	print "Add/Sub      & " + " & ".join([str(i).rjust(5) for i in addSubL]) + " \\\\"
	print "Mul          & " + " & ".join([str(i).rjust(5) for i in mulL])    + " \\\\"
	print "Div          & " + " & ".join([str(i).rjust(5) for i in divL])    + " \\\\"
	print "Exp          & " + " & ".join([str(i).rjust(5) for i in expL])    + " \\\\"
	print "Pow          & " + " & ".join([str(i).rjust(5) for i in powL])    + " \\\\"
	print "Other        & " + " & ".join([str(i).rjust(5) for i in otherL])  + " \\\\"
	print "Total        & " + " & ".join([str(i).rjust(5) for i in totalL])  + " \\\\"
	
	ALML    = []
	DSPL    = []
	CostL   = []
	relALML = []
	relDSPL = []
	for m in [models[m] for m in dmodels]:
		ALML    += [m.optResults[ol].nALM for ol in optLevels]
		relALML += [ofALMs(m.optResults[ol].nALM) for ol in optLevels]
		DSPL    += [m.optResults[ol].nDSP for ol in optLevels]
		relDSPL += [ofDSPs(m.optResults[ol].nDSP) for ol in optLevels]
		CostL   += [m.optResults[ol].cost for ol in optLevels]
	
	print "ALM [\\%]     & " + " & ".join([("%.1f" % f).rjust(5) for f in relALML]) + " \\\\"
	print "DSP [\\%]     & " + " & ".join([("%.1f" % f).rjust(5) for f in relDSPL]) + " \\\\"
	print "Cost (norm.) & " + " & ".join([("%.2f" % (f/CostL[int(i/4)*4])).rjust(5) for i, f in enumerate(CostL)]) + " \\\\"
