/*******************************************************************************
 * Copyright 2015 Julian Oppermann, TU Darmstadt
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

#include "llvm/IR/LLVMContext.h"
#include "llvm/IR/Module.h"
#include "llvm/IR/DataLayout.h"
#include "llvm/ADT/Triple.h"
#include "llvm/Pass.h"
#include "llvm/PassManager.h"
#include "llvm/Transforms/Scalar.h"
#include "llvm/Analysis/Passes.h"
#include "llvm/Analysis/CFGPrinter.h"
#include "llvm/AsmParser/Parser.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/Support/ManagedStatic.h"
#include "llvm/Support/PrettyStackTrace.h"
#include "llvm/Support/Signals.h"
#include "llvm/Support/SourceMgr.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Target/TargetLibraryInfo.h"
#include "llvm/CodeGen/CommandFlags.h"

#include "CustomPasses.h"

#include <string>
#include <cstring>

using namespace llvm;
using namespace cellml::opt;

cl::OptionCategory cellml_opt_options("cellml-opt specific options");

cl::opt<std::string> InputFilename(cl::Positional,
		cl::desc("<input .ll file>"), cl::init("-"), cl::cat(cellml_opt_options));
cl::opt<std::string> ModelName("model-name", cl::Optional,
		cl::desc("Specify the model name"), cl::init("unknown model"), cl::cat(cellml_opt_options));
cl::opt<std::string> EquationFunction("equation-function", cl::Optional,
		cl::desc("Specify the name of the function containing the equation system"), cl::init("computeRates"), cl::cat(cellml_opt_options));
cl::opt<bool> DotCfg("dot-cfg", cl::Optional,
		cl::desc("Dump the equation function's CFG in DOT format"), cl::init(false), cl::cat(cellml_opt_options));
cl::opt<bool> DotEquations("dot-eqs", cl::Optional,
		cl::desc("Dump equation system in DOT format (one graph per equation)"), cl::init(false), cl::cat(cellml_opt_options));

enum OptLevel_t { O0, Oz, OzU, OEqR, OEqC, OzEqC };

cl::opt<OptLevel_t> OptLevel(cl::desc("Choose optimization level:"),
	cl::values(
			clEnumVal(O0,    "No optimisations"),
			clEnumVal(Oz,    "No optimisations (assume -Oz outside of cellml-opt)"),
			clEnumVal(OzU,   "No optimisations (assume -Oz -ffast-math outside of cellml-opt)"),
			clEnumVal(OEqR,  "Enable tailored recipe of built-in FP-safe optimisations"),
			clEnumVal(OEqC,  "Enable tailored unsafe recipe of built-in optimisations and custom transformations"),
			clEnumVal(OzEqC, "Like OEqC, but assume -Oz -ffast-math outside of cellml-opt"),
			clEnumValEnd
	),
	cl::Optional, cl::cat(cellml_opt_options)
);

int main(int argc, char** argv) {
	sys::PrintStackTraceOnErrorSignal();
	PrettyStackTraceProgram X(argc, argv);
	LLVMContext &C = getGlobalContext();
	llvm_shutdown_obj Y;

	StringMap<cl::Option*> Map;
	cl::getRegisteredOptions(Map);
	for (auto& entry : Map) {
		cl::Option *o = entry.second;
		if (o->Category == &cellml_opt_options)
			continue;
		if (strncmp(o->ArgStr, "help", strlen(o->ArgStr)) == 0
		 || strncmp(o->ArgStr, "stats", strlen(o->ArgStr)) == 0)
			continue;
		o->setHiddenFlag(cl::Hidden);
	}

	cl::ParseCommandLineOptions(argc, argv);
	
	SMDiagnostic Err;
	Module *M = ParseAssemblyFile(InputFilename, Err, C);
	if (! M) {
		errs() << "Could not parse module.\n";
		Err.print("cellml-opt", errs());
		return 1;
	}
	
	Function *EF;
	for (Module::iterator MI = M->begin(), ME = M->end(); MI != ME; MI++) {
		if (MI->getName().equals(EquationFunction)) {
			EF = MI;
			break;
		}
	}

	StringRef OptimisationLevelName;
	switch (OptLevel) {
	case O0:    OptimisationLevelName = "0";    break;
	case Oz:    OptimisationLevelName = "z";    break;
	case OzU:   OptimisationLevelName = "zU";   break;
	case OEqR:  OptimisationLevelName = "EqR";  break;
	case OEqC:  OptimisationLevelName = "EqC";  break;
	case OzEqC: OptimisationLevelName = "zEqC"; break;
	}

	FunctionPassManager Prepare(M), Optimise(M), Synthesise(M);

	Triple T(M->getTargetTriple());
	TargetLibraryInfo *TLI = new TargetLibraryInfo(T);
	TLI->disableAllFunctions();
	TLI->setAvailable(LibFunc::pow);
	TLI->setAvailable(LibFunc::exp);
	TLI->setAvailable(LibFunc::log);
	TLI->setAvailable(LibFunc::floor);
	TLI->setAvailable(LibFunc::fabs);

	TargetLibraryInfo *TLIOpt = new TargetLibraryInfo(*TLI); // copy

	Prepare.add(new DataLayoutPass(M));
	Prepare.add(TLI);
	Prepare.add(createPromoteMemoryToRegisterPass());
	Prepare.add(createIntrinsicMapperPass());
	if (OptLevel >= OEqR)
		Prepare.add(createNoAliasInserterPass());

	Prepare.doInitialization();
	Prepare.run(*EF);
	Prepare.doFinalization();

	if (OptLevel >= OEqR) {
		Optimise.add(new DataLayoutPass(M));
		Optimise.add(TLIOpt);
		Optimise.add(createInstanceInfoPass(ModelName, OptimisationLevelName));
		Optimise.add(createBasicAliasAnalysisPass());
		Optimise.add(createCFGSimplificationPass());
		Optimise.add(createInstructionCombiningPass());
		Optimise.add(createGVNPass());
		if (OptLevel >= OEqC)
			Optimise.add(createOptimisationPlaygroundPass());
		Optimise.add(createDeadCodeEliminationPass());

		Optimise.doInitialization();
		for (int i = 0; i < 10 && Optimise.run(*EF); i++);
		Optimise.doFinalization();
	}

	Synthesise.add(createInstanceInfoPass(ModelName, OptimisationLevelName));
	if (DotCfg)
		Synthesise.add(createCFGPrinterPass());
	Synthesise.add(createEquationSystemBuilderPass());
	Synthesise.add(createExpressionReuserPass());
	Synthesise.add(createCSynthesiserPass());
	Synthesise.add(createAreaEstimaterPass());
	if (DotEquations)
		Synthesise.add(createDotWriterPass());

	Synthesise.doInitialization();
	Synthesise.run(*EF);
	Synthesise.doFinalization();

	return 0;
}
