/*******************************************************************************
 * Copyright 2015 Ting Yu, Julian Oppermann
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *******************************************************************************
 *
 * Generic driver performing a numerical integration as it would be done during
 * the simulation of a CellML model.
 *
 ******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <math.h>

#define HEX2FLOAT(X)    (float)(*(float*)&X)
#define FLOAT2HEX(X)    (int)(*(int*)&X)

#include "model.h"

#define INPUT_WIDTH     (NUM_STATES + NUM_CONSTANTS)
#define OUTPUT_WIDTH    (NUM_RATES  + NUM_ALGEBRAIC)

#define CELL_COUNT      10

extern void initConsts(float* CONSTANTS, float* RATES, float *STATES);
extern void computeRates(float VOI, float* CONSTANTS, float* RATES, float* STATES, float* ALGEBRAIC);

void softwareCompute(float *buff_in, float *buff_out)
{
	int i, j, z;
	float VOI, STATES[NUM_STATES], RATES[NUM_RATES], CONSTANTS[NUM_CONSTANTS], ALGEBRAIC[NUM_ALGEBRAIC];
	
	memset(STATES,    0, NUM_STATES    * sizeof(float));
	memset(RATES,     0, NUM_RATES     * sizeof(float));
	memset(CONSTANTS, 0, NUM_CONSTANTS * sizeof(float));
	memset(ALGEBRAIC, 0, NUM_ALGEBRAIC * sizeof(float));
	
	for (i = 0; i < CELL_COUNT; i++) {
		VOI = 1.0f;
		memcpy(STATES,    buff_in + (i * INPUT_WIDTH),              sizeof(STATES));
		memcpy(CONSTANTS, buff_in + (i * INPUT_WIDTH + NUM_STATES), sizeof(CONSTANTS));
		
		for (j = 0; j < 1000; j++) {
			computeRates(VOI, CONSTANTS, RATES, STATES, ALGEBRAIC);
			for (z = 0; z < NUM_STATES; z++) {
				STATES[z] += RATES[z] * 1e-6f;
			}
			VOI += 1e-6f;
		}
		
		memcpy(buff_out + i * OUTPUT_WIDTH,             RATES,     sizeof(RATES));
		memcpy(buff_out + i * OUTPUT_WIDTH + NUM_RATES, ALGEBRAIC, sizeof(ALGEBRAIC));	
	}
}

int main()
{
	int i,j;
	
	float buff_in[CELL_COUNT * INPUT_WIDTH], buff_out[CELL_COUNT * OUTPUT_WIDTH];
	float STATES[NUM_STATES], RATES[NUM_RATES], CONSTANTS[NUM_CONSTANTS];
	
	memset(buff_in,   0, CELL_COUNT * INPUT_WIDTH  * sizeof(float));
	memset(buff_out,  0, CELL_COUNT * OUTPUT_WIDTH * sizeof(float));
	memset(STATES,    0, NUM_STATES                * sizeof(float));
	memset(RATES,     0, NUM_RATES                 * sizeof(float));
	memset(CONSTANTS, 0, NUM_CONSTANTS             * sizeof(float));
	
	for (i = 0; i < CELL_COUNT; i++){
		initConsts(CONSTANTS, RATES, STATES);
		STATES[0] += i / 1000.0f;
		memcpy(buff_in + (i * INPUT_WIDTH)             , STATES,    sizeof(STATES));
		memcpy(buff_in + (i * INPUT_WIDTH + NUM_STATES), CONSTANTS, sizeof(CONSTANTS));
	}
	
	struct timeval base_time;
	gettimeofday(&base_time, NULL);
	
	for (i = 0; i < 100; i++){
		softwareCompute(buff_in, buff_out);
	}
	
	struct timeval curr_time;
	gettimeofday(&curr_time, NULL);
	
	long microseconds = curr_time.tv_usec - base_time.tv_usec + 1000000 * (curr_time.tv_sec - base_time.tv_sec);
	
	printf("Processing %d cells took %ld microseconds\n", CELL_COUNT, microseconds);
	
	// print the values
	printf("Results :\n");
	for (i = 0; i < CELL_COUNT; i++) {
		for (j = 0; j < OUTPUT_WIDTH; j++)
			printf("%.9f\n", buff_out[i * OUTPUT_WIDTH + j]);
	}
	
	return 0;
}
